/*
	ttyserver -- MDQS server for generic device on tty port
				(supports DC3/DC1 flow control)

	last edit:	16-May-1986	D A Gwyn

	Absolutely no interpretation of the data is done here,
	although the terminal handler may be asked to perform a
	few standard mappings; if any other transformation is
	needed, do it before spooling the data.

	Intended use is for plotters, laser printers, etc.

	Line count limit is meaningless and therefore ignored.

	Unless the "noff" line-printer option is specified,
	a form-feed will be output after each copy.

	The "banner" option enables headers and trailers.
	Don't specify both "banner" and "noff" at the same time.

	If the "crmod" option is specified, newline characters
	will be output as CR and LF, in that order.

	If the "xtabs" option is specified, horizontal tab
	characters will be converted to sequences of spaces.

	The "speed=" and "parity=" options are also supported.
	"parity=none" requests an 8-bit wide data path.  Some
	systems offer better support for this than do others;
	in particular, this may preclude "crmod" and "xtabs".

	This code can be compiled under any UNIX environment;
	it will figure out the correct conditional compilation.
*/

#include	<pwd.h>
#include	<signal.h>
#include	<stdio.h>

#include	"queue.h"
#include	"qreplys.h"

typedef int	bool;			/* Boolean data type */
#define	false	0
#define	true	1
#define BELL	007
#define PAGE	66

extern void		exit(), free();
extern char		*ctime(), *malloc(), *strcpy();
extern int		atoi(), chdir(), close(), ioctl(), open(),
			read(), strcmp(), strlen(), strncmp(), unlink(),
			write();
extern long		lseek(), time();
extern struct passwd	*getpwuid();

extern void		doflush(), dorestart(), prheader(), prtrailer(),
			retmsg(), statmsg();
extern struct request	*gethdr();

extern char		*Qdatadir;	/* queue file directory name */

static void		arg_init(), end_tty(), fatal(), set_baud(),
			set_crmod(), set_xtabs(), set_parity(),
			set_tty(), write_err(), x_option();
static bool		print();
void			err();

static struct request	r;		/* unpacked user request data */
static int		ofd;		/* stdout file descriptor */
static int		lpoptions = 0;	/* line-printer style options */
static bool		banner = false;	/* "permit headers/trailers" */
static bool		dev_ok = true;	/* "no write errors yet" */
static bool		tty_inited = false;	/* "port initialized" */

/* The following are used by prheader() and prtrailer(): */
char			source[LINESIZ];/* source of request */
char			*title = NULL;	/* user-specified title */
int			numhdrs = 1;	/* desired number of headers */
int			numtrailers = 0;/* desired number of trailers */
int			numcols = 132;	/* width of printer */
long			linecount = 0L;	/* number of lines printed */
long			timenow;	/* current time of day */
int			single =0;

/*
	As an MDQS queue server, this process will be started up
	with the queue control file as its standard input
	and the device of interest open on its standard output.
*/

main( argc, argv )
	int	argc;
	char	*argv[];
	{
	(void)signal( SIGHUP, SIG_IGN );	/* paranoid */
	(void)signal( SIGEMT, dorestart );
	(void)signal( SIGTERM, doflush );

	arg_init( argc, argv );		/* process option arguments */

	timenow = time( (long *)NULL );

	setbuf( stdout, malloc( BUFSIZ ) );
	ofd = fileno( stdout );

	if ( gethdr( stdin, &r ) == NULL )
		fatal( "error reading request header" );

	{				/* process control lines */
	char	linebuf[LINESIZ];

	while ( fgets( linebuf, (int)sizeof linebuf, stdin ) != NULL )
		{
		register char	*cp;

		if ( *(cp = &linebuf[strlen( linebuf ) - 1]) == '\n' )
			*cp = '\0';

		cp = &linebuf[0];
		switch ( *cp++ )
			{
		case 'B':		/* print and remove banner */
			{
			char	msg[LINESIZ];	/* status message */

			(void)sprintf( msg, "printing banner %s", cp );
			statmsg( msg );
			}
			{
			char	dfile[LINESIZ];	/* data file name */

			(void)sprintf( dfile, "%s/%s", Qdatadir, cp );

			if ( banner && dev_ok )
				{
				set_tty();
				prheader( dfile, &r );
				}

			if ( unlink( dfile ) != 0 )
				err( "couldn't unlink banner" );
			}
			break;

		case 'D':		/* change working directory */
			if ( chdir( cp ) != 0 )
				{
				err( "chdir error" );
				(void)chdir( "/tmp" );	/* safe */
				}
			break;

		case 'F':		/* print user file */
			{
			char	msg[LINESIZ];	/* status message */

			(void)sprintf( msg, "outputting %s", cp );
			statmsg( msg );
			}

			set_tty();

			if ( dev_ok
			  && print( cp, r.r_copies )
			  && (lpoptions & NOKEEP) != 0
			  && unlink( cp ) != 0
			   )
				err( "couldn't unlink data file" );
			break;

		case 'H':		/* print header */
			{
			char	msg[LINESIZ];	/* status message */

			(void)sprintf( msg, "printing header %s", cp );
			statmsg( msg );
			}
			if ( banner && dev_ok )
				{
				set_tty();
				prheader( cp, &r );
				}
			break;

		case 'I':		/* print spooled file */
			{
			char	msg[LINESIZ];	/* status message */

			(void)sprintf( msg, "outputting %s",
					title != NULL ? title : cp
				     );
			statmsg( msg );
			}
			{
			char	dfile[LINESIZ];	/* data file name */

			(void)sprintf( dfile, "%s/%s", Qdatadir, cp );

			if ( dev_ok )
				{
				set_tty();
				(void)print( dfile, r.r_copies );
				}

			if ( unlink( dfile ) != 0 )
				err( "couldn't unlink spooled file" );
			}
			break;

		case 'O':		/* line-printer style options */
			lpoptions |= atoi( cp );
			break;

		case 'S':
			single++;
			break;

		case 'T':		/* title */
			if ( title != NULL )
				free( title );

			title = malloc( strlen( cp ) + 1 );
			if ( title != NULL )
				(void)strcpy( title, cp );
			break;

		case 'U':		/* user name */
			(void)strcpy( source, cp );
			break;

		case 'X':		/* extended options */
			x_option( cp );
			break;

		default:		/* ignore unknown controls */
			break;
			}
		}
	}

	if ( banner && dev_ok && numtrailers > 0 )
		{
		set_tty();
		prtrailer( &r );
		}

	if ( dev_ok && fflush( stdout ) == EOF )
		write_err();

	end_tty();

	retmsg( RP_OK, (char *)0 );
#ifdef	lint
	return 0;
#else
	exit( 0 );
#endif
	}


/*
	Process line-printer style option arguments.
*/

static void
arg_init( argc, argv )
	register int	argc;		/* argument count */
	register char	*argv[];	/* argument strings */
	{
	while ( ++argv, --argc > 0 )
		if ( strcmp( argv[0], "banner" ) == 0 )
			banner = true;
		else if ( strncmp( argv[0], "columns=", 8 ) == 0 )
			{
			if ( (numcols = atoi( &argv[0][8] )) <= 0 )
				fatal( "bad columns= option value" );
			}
		else if ( strcmp( argv[0], "crmod" ) == 0 )
			set_crmod();
		else if ( strncmp( argv[0], "nheaders=", 9 ) == 0 )
			{
			if ( (numhdrs = atoi( &argv[0][9] )) < 0 )
				fatal( "bad nheaders= option value" );
			}
		else if ( strcmp( argv[0], "noff" ) == 0 )
			lpoptions |= NOFF;	/* "no form-feeds" */
		else if ( strncmp( argv[0], "ntrailers=", 10 ) == 0 )
			{
			if ( (numtrailers = atoi( &argv[0][10] )) < 0 )
				fatal( "bad ntrailers= option value" );
			}
		else if ( strncmp( argv[0], "parity=", 7 ) == 0 )
			set_parity( &argv[0][7] );
		else if ( strncmp( argv[0], "speed=", 6 ) == 0 )
			set_baud( atoi( &argv[0][6] ) );
		else if ( strcmp( argv[0], "xtabs" ) == 0 )
			set_xtabs();
		/* else ignore option */
	}


/*
	Process user-specified extended option.

	Only those that are "obviously" safe are permitted.
*/

static void
x_option( s )
	register char	*s;		/* option string */
	{
	if ( strcmp( s, "crmod" ) == 0 )
		set_crmod();
	else if ( strcmp( s, "noff" ) == 0 )
		lpoptions |= NOFF;	/* "no form-feeds" */
	else if ( strcmp( s, "xtabs" ) == 0 )
		set_xtabs();
/*	else if ( strcmp( s, "single" ) == 0 )
 *		set_single();		
 */
	/* else ignore option */
	}


/*
	Dump specified file onto standard output without interpretation.
*/

static bool
print( file, cnt )			/* returns true if successful */
	char		*file;		/* name of file */
	register int	cnt;		/* number of copies */
	{
	register int	ifd;		/* input file descriptor */

	if ( (ifd = open( file, 0 )) < 0 )	/* O_RDONLY */
		{
		if ( banner && dev_ok )
			{
			(void)printf(
			  "\n\n\n*** couldn't open file \"%s\" ***\r\f",
				      file
				    );
			/* Berkeley doesn't say what printf() returns */
			dev_ok = ferror( stdout ) == 0
			      && fflush( stdout ) != EOF;
			}
		else
			err( "couldn't open data file" );

		return false;
		}
	else	{
		register bool	ok = dev_ok;	/* false on error */

		/* Flush any output already buffered for stdout. */
		if ( ok
		  && !(ok = fflush( stdout ) != EOF)
		   )
			write_err();

		while ( ok && --cnt >= 0 )
			{
			ok = copy_out( ifd );	/* print one copy */

			/* Print a form-feed after each
			   copy, unless otherwise specified. */

			if ( (lpoptions & NOFF) == 0
			  && dev_ok
			  && ( putchar( '\f' ) == EOF
			    || fflush( stdout ) == EOF
			     )
			   )
				ok = dev_ok = false;

			if ( cnt > 0		/* prepare for read */
			  && ok
			  && !(ok = lseek( ifd, 0L, 0 ) == 0L)
			   )
				err( "couldn't rewind file" );
			}

		(void)close( ifd );
		return ok;
		}
	}


/*
	Print copy of stream to standard output without interpretation.

	This bypasses stdio in the name of the god of efficiency.

	It also attempts to support any kernel efficiency hacks for
	avoidance of read/write data copying.  I don't know of very
	many implementations that will benefit from this, though.
*/

static bool
copy_out( ifd )				/* returns true if successful */
	register int	ifd;		/* input file descriptor */
	{
	char		buf[2 * BUFSIZ - 1];	/* i/o buffer + slop */
	register int	n;		/* i/o byte count */
	register char	*bp;		/* -> optimally-aligned block */
	register int  	lncnt=0;
	register char   ch;

	/* If the following gives you problems, just set
	   bp = buf and to hell with the efficiency god. */
	if ( (long)(bp = buf) % BUFSIZ != 0L )
		bp = (char *)((long)bp + BUFSIZ - (long)buf % BUFSIZ);

	/* Roll the file through as efficiently as possible. */

	while ( (n = read( ifd, bp, BUFSIZ )) > 0 ) {
		if (lncnt == PAGE){
			putchar(BELL);
			lncnt = 0;
			while  (ch=getc(stdout) != '\n'){
				sleep (60);
				putchar(BELL);				
			}
		}
		if ( write( ofd, bp, n ) != n )
			{
			write_err();
			return false;
			}
		if (single)
			lncnt++;
	}
	if ( n < 0 )
		{
		err( "data file read failed" );
		return false;
		}
	else
		return true;		/* success */
	}


/*
	Print write-error message on standard error output.
*/

static void
write_err()
	{
	err( "device write error" );

	dev_ok = false;			/* avoid future writes */
	}


/*
	Print error message on standard error output.

	Banner() invokes this routine with printf-like parameters,
	which we ignore in order to avoid portability issues, alas.
*/

/*VARARGS1*/
void
err( msg )
	char	*msg;
	{
	(void)fputs( "ttyserver: ", stderr );
	(void)fputs( msg, stderr );
	(void)putc( '\n', stderr );
	}


/*
	Print fatal error message on standard error output, then die.
*/

static void
fatal( msg )
	char	*msg;
	{
	end_tty();

	err( msg );
	retmsg( RP_FATAL, msg );
	exit( 10 );
	}


/*
	UNIX TERMINAL HANDLER INTERFACE
*/


#include	<sgtty.h>
#ifndef	TANDEM	/* USG derivatives */
/* Dammit. */
#undef	NL1
#undef	CR1
#undef	CR2
#undef	CR3
#undef	TAB1
#undef	BS1
#undef	FF1
#undef	B50
#undef	B75
#undef	B110
#undef	B134
#undef	B150
#undef	B200
#undef	B300
#undef	B600
#undef	B1200
#undef	B1800
#undef	B2400
#undef	B4800
#undef	B9600
#undef	EXTA
#undef	EXTB
#undef	HUPCL
#undef	ECHO
#include	<termio.h>
#endif

#ifdef	TANDEM	/* 7th Edition derivatives */
#ifndef	XTABS
#define	XTABS	(TAB1 | TAB2)
#endif
#ifdef	LLITOUT
static bool	litout = false;		/* "suppress translations" */
#endif
static struct sgttyb ttyb =	/* tty line setup commands */
	{
	B9600,				/* sg_ispeed */
	B9600,				/* sg_ospeed */
	0,				/* sg_erase */
	0,				/* sg_kill */
	EVENP | ODDP			/* sg_flags */
#ifdef	CBREAK	/* can avoid special treatment of EOT */
		     | CBREAK		/* sg_flags, continued */
#endif
	};
#else		/* USG derivatives */
static struct termio ttyb =	/* tty line setup commands */
	{
	BRKINT | IGNCR | ISTRIP | IXON,	/* c_iflag */
	0,				/* c_oflag */
	B9600 | CS8 | CREAD | CLOCAL,	/* c_cflag */
	0,				/* c_lflag */
	0,				/* c_line */
	{				/* c_cc[] */
	CDEL,				/* INTR */
	CDEL,				/* QUIT */
	CDEL,				/* ERASE */
	CDEL,				/* KILL */
	CNUL,				/* EOF or MIN */
	CNUL,				/* EOL or TIME */
	CNUL,				/* reserved */
	CNUL				/* SWTCH */
	}
	};
#endif


/*
	Override default baud rate setting.
*/

static void
set_baud( baud )
	int	baud;
	{
	static struct rate
		{
		int	spec;		/* user specification */
		int	speed;		/* bits to set */
		}		tbl[] =
	{
		{	50,	B50	},
		{	75,	B75	},
		{	110,	B110	},
		{	134,	B134	},
		{	150,	B150	},
		{	200,	B200	},
		{	300,	B300	},
		{	600,	B600	},
		{	1200,	B1200	},
		{	1800,	B1800	},
		{	2400,	B2400	},
		{	4800,	B4800	},
		{	9600,	B9600	},
		{	19200,	EXTA	},
		{	38400,	EXTB	}	/* even on PDP-11? */
	};
	register struct rate	*tp;

	for ( tp = tbl; tp < &tbl[sizeof tbl / sizeof tbl[0]]; ++tp )
		if ( baud == tp->spec )
			break;

	if ( tp < &tbl[sizeof tbl / sizeof tbl[0]] )
		{
#ifdef	TANDEM	/* 7th Edition derivatives */
		ttyb.sg_ispeed = ttyb.sg_ospeed = tp->speed;
#else		/* USG derivatives */
		ttyb.c_cflag &= ~CBAUD;	/* clear baud-rate field */
		ttyb.c_cflag |= tp->speed;
		if ( baud <= 134 )
			ttyb.c_cflag |= CSTOPB;
#endif
		}
	else
		fatal( "bad speed= option value" );
	}


/*
	Override default parity setting.
*/

static void
set_parity( parity )
	char	*parity;		/* user parity specification */
	{
	static struct par
		{
		char	*spec;		/* user specification */
		int	bits;		/* bits to set */
#ifdef	TANDEM
#ifdef	LLITOUT
		bool	use_litout;	/* use LITOUT mode */
#endif
#endif
		}		tbl[] =
	{
#ifdef	TANDEM	/* 7th Edition derivatives */
#ifdef	LLITOUT
		{	"even",		EVENP,		false	},
		{	"odd",		ODDP,		false	},
		{	"mark",		EVENP | ODDP,	false	},
		{	"space",	0,		false	},
		{	"any",		EVENP | ODDP,	false	},
		{	"none",		EVENP | ODDP,	true	}
#else
		{	"even",		EVENP			},
		{	"odd",		ODDP			},
		{	"mark",		EVENP | ODDP		},
		{	"space",	0			},
		{	"any",		EVENP | ODDP		},
		{	"none",		EVENP | ODDP		}
#endif
#else		/* USG derivatives */
		{	"even",		CS7 | PARENB		},
		{	"odd",		CS7 | PARENB | PARODD	},
		{	"mark",		CS7 | CSTOPB		},
		{	"space",	CS8			},
		{	"any",		CS8			},
		{	"none",		CS8			}
#endif
	};
	register struct par	*tp;

	for ( tp = tbl; tp < &tbl[sizeof tbl / sizeof tbl[0]]; ++tp )
		if ( strcmp( parity, tp->spec ) == 0 )
			break;

	if ( tp < &tbl[sizeof tbl / sizeof tbl[0]] )
		{
#ifdef	TANDEM	/* 7th Edition derivatives */
#ifdef	LLITOUT
		litout = tp->use_litout;
#endif
		ttyb.sg_flags &= ~(EVENP | ODDP);
						/* clear parity field */
		ttyb.sg_flags |= tp->bits;
#else		/* USG derivatives */
		ttyb.c_cflag &= ~(CSIZE | PARENB | PARODD);
						/* clear parity field */
		ttyb.c_cflag = tp->bits;
#endif
		}
	else
		fatal( "bad parity= option value" );
	}


/*
	Enable NL -> (CR,LF) mapping.
*/

static void
set_crmod()
	{
#ifdef	TANDEM	/* 7th Edition derivatives */
	ttyb.sg_flags |= CRMOD;
#else		/* USG derivatives */
	ttyb.c_oflag |= OPOST | ONLCR;
#endif
	}


/*
	Enable tab expansion.
*/

static void
set_xtabs()
	{
#ifdef	TANDEM	/* 7th Edition derivatives */
	ttyb.sg_flags |= XTABS;
#else		/* USG derivatives */
	ttyb.c_oflag |= OPOST | TAB3;
#endif
	}


/*
	Establish desired terminal handler state on open port.
*/

static void
set_tty()
	{
	if ( tty_inited )
		return;			/* already set up */
	tty_inited = true;		/* set once-only flag */

	/* Attempt to keep other processes from interfering.
	   If the spooling system is properly configured,
	   this should not be necessary. */

#ifdef	TIOCEXCL
	(void)ioctl( ofd, TIOCEXCL, (int *)0 );	/* exclusive use */
#endif

	/* Establish communications by raising DTR, if so equipped.
	   This should be done by the tty handler when opening
	   the port, but we might as well be on the safe side. */

#ifdef	TIOCSDTR
	(void)ioctl( ofd, TIOCSDTR, (int *)0 );	/* set DTR on port */
#endif

	/* Flush any data left over from some other process.
	   This cannot be ttyserver data,
	   since we wait for ours to drain before terminating. */

#ifdef	TCFLSH
	(void)ioctl( ofd, TCFLSH, 2 );		/* flush i/o queues */
#else
#ifdef	TIOCFLUSH
	{
	static int	lflush = 3;

	(void)ioctl( ofd, TIOCFLUSH, &lflush );	/* flush i/o queues */
	}
#endif
#endif

	/* Set desired port parameters and disable special chars. */

#ifdef	TIOCSETD
	{
	static int	ndisc = NTTYDISC;
	int		odisc;

	if ( ioctl( ofd, TIOCGETD, &odisc ) == 0 && odisc != ndisc )
		/* use new tty handler */
		(void)ioctl( ofd, TIOCSETD, &ndisc );
	}
#endif

#ifdef	TANDEM	/* 7th Edition derivatives */
	if ( stty( ofd, &ttyb ) != 0 )
#else
	if ( ioctl( ofd, TCSETAF, &ttyb ) != 0 )
#endif
		fatal( "couldn't set tty modes" );

#ifdef	TIOCLSET
	{
	static int	lflags = LDECCTQ;

#ifdef	LLITOUT
	if ( litout )
		lflags |= LLITOUT;
#endif
	if ( ioctl( ofd, TIOCLSET, &lflags ) != 0 )	/* local mode */
		err( "couldn't set local flags" );
#ifdef	LLITOUT
	if ( litout )
		{
		/* The following 4.2BSD hack courtesy of Chris Torek. */
		(void)gtty( ofd, &ttyb );
		(void)ioctl( ofd, TIOCSETN, &ttyb );	/* make stick */
		}
#endif
	}
#endif

#ifdef	TIOCSETC
	{
	static struct tchars	tc =
		{
		(char)-1,		/* t_intrc */
		(char)-1,		/* t_quitc */
		0x11,			/* t_startc = DC1 */
		0x13,			/* t_stopc = DC3 */
		(char)-1,		/* t_eofc */
		(char)-1		/* t_brkc */
		};

	(void)ioctl( ofd, TIOCSETC, &tc );	/* set special chars */
	}
#endif

#ifdef	TIOCSLTC
	{
	static struct ltchars	ltc =
		{
		(char)-1,		/* t_suspc */
		(char)-1,		/* t_dsuspc */
		(char)-1,		/* t_rprntc */
		(char)-1,		/* t_flushc */
		(char)-1,		/* t_werasc */
		(char)-1		/* t_lnextc */
		};

	(void)ioctl( ofd, TIOCSLTC, &ltc );	/* set local specials */
	}
#endif

	/* Just in case output is stalled, resume.
	   This is not a very good strategy,
	   especially for devices like HP plotters,
	   but the tty handler messes up things
	   whenever the port is newly opened,
	   so we might have missed a DC1.
	   A correct solution would require
	   leaving the port open on a server
	   and passing data to it for output,
	   rather than starting up a fresh server. */

#ifdef	TCXONC
	(void)ioctl( ofd, TCXONC, 1 );		/* start output */
#else
#ifdef	TIOCSTART
	(void)ioctl( ofd, TIOCSTART, (int *)0 );	/* ditto */
#endif
#endif
	}


/*
	Fix up terminal handler state on open port before terminating.
*/

static void
end_tty()
	{
	if ( !tty_inited )
		return;			/* state hasn't been changed */

	/* Wait for output to drain (side-effect). */

#ifdef	TANDEM	/* 7th Edition derivatives */
	(void)stty( ofd, &ttyb );
#else
	(void)ioctl( ofd, TCSETAW, &ttyb );
#endif

	/* Flush input that may have been generated by the device. */

#ifdef	TCFLSH
	(void)ioctl( ofd, TCFLSH, 0 );	/* flush input queue */
#else
#ifdef	TIOCFLUSH
	{
	static int	lflush = 1;

	(void)ioctl( ofd, TIOCFLUSH, &lflush );	/* flush input queue */
	}
#endif
#endif

	/* We could drop DTR here, but why bother. */

	/* Allow shared access to port (safety net, not required). */

#ifdef	TIOCNXCL
	(void)ioctl( ofd, TIOCNXCL, (int *)0 );	/* non-exclusive use */
#endif
	}

/*
	Single forms option
*/

static void
set_single ()
	{
	single++;
	}
